#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "network.h"

#define UDP_BUFFER_SIZE  1024
#define TCP_BUFFER_SIZE 16000
#define SA struct sockaddr

TCPLstn *
tcp_open_listener (int port)
{
  TCPLstn     *lstn;
  int         res;

  /* malloc the tcplstn dude into existence */
  if (NULL == (lstn = malloc (sizeof(struct _TCPLstn))) ) {
    fprintf(stderr, "tcp_open_listener: failed in malloc\n");
    return NULL;
  }
  
  /* initialize socket info stuff */
  lstn->port = port;
  lstn->sock_info.sin_family = AF_INET;
  lstn->sock_info.sin_addr.s_addr = htonl (INADDR_ANY);
  lstn->sock_info.sin_port = htons (port);

  /* create the socket */
  if (0 > (lstn->sock = socket (AF_INET, SOCK_STREAM, 0)) ) {
    fprintf (stderr, "tcp_open_listener: failed to create socket for port %d\n", port);
    tcp_close_listener (lstn);
    return NULL;
  }

  /* bind to the specified interface and port */
  if (0 > (res = bind (lstn->sock, (struct sockaddr *)&lstn->sock_info, sizeof(lstn->sock_info) )) ) {
    fprintf (stderr, "tcp_open_listener: failed to bind socket to port %d\n", port);
    lstn->sock = -1;
    tcp_close_listener (lstn);
    return NULL;
  }

  /* start the listener */
  if (0 > (res = listen (lstn->sock, 64)) ) {
    fprintf (stderr, "tcp_open_listener: failed to open listener on port %d\n", port);
    tcp_close_listener (lstn);
    return NULL;
  }

  printf ("tcp_open_listener: done [%d]\n", lstn->port);

  return lstn;
}

void
tcp_close_listener (TCPLstn *lstn)
{
  int     port;

  if (lstn == NULL)
    return;

  port = lstn->port;

  if (lstn->sock > 0 ) {
    if (0 > shutdown (lstn->sock, SHUT_RDWR))
      fprintf (stderr, "tcp_close_listener: failed to shutdown listener [%d]\n", lstn->port);
    else
      close (lstn->sock);
  }

  free (lstn);

  printf ("tcp_close_listener: done [%d]\n", port);
}

TCPConn *
tcp_open_connection (char *addr, int port)
{
  TCPConn     *conn;
  int         res;

  if (addr == NULL || port < 1)
    return NULL;

  if (NULL == (conn = malloc (sizeof(struct _TCPConn))) ) {
    fprintf (stderr, "tcp_open_connection: failed in malloc [1]\n");
    return NULL;
  }

  if (NULL == (conn->buf = malloc (sizeof(char) * TCP_BUFFER_SIZE)) ) {
    fprintf (stderr, "tcp_open_connection: failed in malloc [2]\n");
    tcp_close_connection (conn);
    return NULL;
  }

  if (NULL == (conn->sock_info = malloc (sizeof(struct sockaddr_in))) ) {
    fprintf (stderr, "tcp_open_connection: failed in malloc [3]\n");
    tcp_close_connection (conn);
    return NULL;
  }
  printf ("tcp_open_connection conn->sock_info = %p\n", conn->sock_info);

  /* initialize socket info stuff */

  bzero (conn->buf, sizeof(char) * TCP_BUFFER_SIZE);
  bzero (conn->sock_info, sizeof(struct sockaddr_in));
  inet_pton (AF_INET, addr, &conn->sock_info->sin_addr);

  conn->sock_info->sin_family = AF_INET;
  conn->sock_info->sin_port = htons (port);
  conn->is_connected = 0;
  conn->buf_leng = 0;
  conn->addr = NULL;
  conn->bytes_in = 0;
  conn->bytes_out = 0;

  if (0 > (res = socket (AF_INET, SOCK_STREAM, 0)) ) {
    fprintf (stderr, "tcp_open_connection: failed to socket [%s:%d]\n", addr, port);
    tcp_close_connection (conn);
    return NULL;
  }

  conn->sock = res;

  if (0 > (res = connect (conn->sock, (SA *)conn->sock_info, sizeof(struct sockaddr_in))) ) {
    fprintf (stderr, "tcp_open_connection: failed to connect [%s:%d]\n", addr, port);
    tcp_close_connection (conn);
    return NULL;
  }

  conn->is_connected = 1;

  /* printf ("tcp_open_connection: connected to %s:%d\n", addr, port); */

  return conn;
}

UDPConn *
udp_open_connection (char *addr, int port)
{
  UDPConn     *conn;
  int         res;
  const int   on = 1;

  /* malloc the udp conn dude into existence */
  if (NULL == (conn = malloc (sizeof(struct _UDPConn))) ) {
    fprintf (stderr, "udp_open_connection: failed in malloc [1]\n");
    return NULL;
  }
  
  /* malloc the socket read buffer */
  if (NULL == (conn->buf = malloc (sizeof(char) * UDP_BUFFER_SIZE)) ) {
    fprintf (stderr, "udp_open_connection: failed in malloc [2]\n");
    udp_close_connection (conn);
    return NULL;
  }

  /* initialize socket info stuff */

  if (addr[0] == '\0')
    conn->sock_info.sin_addr.s_addr = htonl (INADDR_ANY);
  else
    inet_pton (AF_INET, addr, &conn->sock_info.sin_addr);

  conn->port = port;
  conn->bytes_in = 0;
  conn->bytes_out = 0;
  conn->so_timeout = 0;
  conn->sock_info.sin_family = AF_INET;
  conn->sock_info.sin_port = htons (port);

  /* create the socket */
  if (0 > (conn->sock = socket (AF_INET, SOCK_DGRAM, 0)) ) {
    fprintf (stderr, "udp_open_connection: failed to cerate socket, %s\n", strerror(conn->sock));
    udp_close_connection (conn);
    return NULL;
  }

  /* bind to the specified interface and port */
  if (0 > (res = bind (conn->sock, (struct sockaddr *)&conn->sock_info, sizeof(conn->sock_info) )) ) {
    fprintf (stderr, "udp_open_connection: failed to bind socket to port %d, %s\n", port, strerror(res));
    udp_close_connection (conn);
    return NULL;
  }

  if (0 > (res = setsockopt (conn->sock, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on))) ) {
    fprintf (stderr, "upd_open_connection: error in setsockopt(), %s\n", strerror(errno));
    udp_close_connection (conn);
    return NULL;
  }

  printf ("udp_open_connection: done [%d]\n", conn->port);

  return conn;
}

void
udp_close_connection (UDPConn *conn)
{
  int     port;

  if (conn == NULL)
    return;

  port = conn->port;

  if (conn->sock > 0 )
    if (0 > close (conn->sock))
      fprintf (stderr, "udp_close_connection: failed to close connection [%d]\n", conn->port);

  if (conn->buf != NULL)
    free (conn->buf);

  free (conn);

  printf ("udp_close_connection: done [%d]\n", port);
}

TCPConn *
tcp_accept (TCPLstn *lstn)
{
  TCPConn         *conn;
  socklen_t       socklen;
  char            str[128];

  if (lstn == NULL)
    return NULL;

  if (NULL == (conn = malloc (sizeof(struct _TCPConn))) ) {
    fprintf (stderr, "tcp_accept: failed to malloc a conn\n");
    return NULL;
  }

  conn->sock_info = malloc (sizeof(struct sockaddr_in));
  conn->bytes_in  = 0;
  conn->bytes_out = 0;
  conn->buf_leng  = 0;

  if (0 > (conn->sock = accept(lstn->sock, (struct sockaddr *) conn->sock_info, &socklen)) ) {
    fprintf (stderr, "tcp_accept: failed while doing accept(), %s\n", strerror(conn->sock));
    tcp_close_connection (conn);
    return NULL;
  }

  conn->port = ntohs (conn->sock_info->sin_port);
  
  if (NULL == (inet_ntop (AF_INET, &conn->sock_info->sin_addr, str, sizeof(str))) ) {
    fprintf (stderr, "tcp_accept: error while parsing client address\n");
    tcp_close_connection (conn);
    return NULL;
  }

  conn->addr = malloc (strlen(str) + 1);
  strcpy (conn->addr, str);

  if (NULL == (conn->buf = malloc (sizeof(char) * TCP_BUFFER_SIZE)) ) {
    fprintf (stderr, "tcp_accept: failed in malloc [2]\n");
    tcp_close_connection (conn);
    return NULL;
  }

  bzero (conn->buf, sizeof(char) * TCP_BUFFER_SIZE);
  conn->is_connected = 1;

  printf ("tcp_accept: accepted new connection from %s:%d\n", conn->addr, conn->port);

  return conn;
}

void
tcp_close_connection (TCPConn *conn)
{
  int     port, res;

  if (conn == NULL)
    return;

  port = conn->port;

  if (conn->sock > 0 ) {
    if (0 > (res = shutdown (conn->sock, SHUT_RDWR)))
      fprintf (stderr, "tcp_close_connection: failed to shutdown connection [%d]\n", conn->port);
    else
      close (conn->sock);
  }

  if (conn->sock_info != NULL) {
    printf ("tcp_close_connection conn->sock_info = %p\n", conn->sock_info);
    free (conn->sock_info);
  }

  if (conn->buf != NULL)
    free (conn->buf);

  if (conn->addr != NULL)
    free (conn->addr);

  free (conn);

  printf ("tcp_close_connection: done [%d]\n", port);
}

int
tcp_send (TCPConn *conn, GEvent *gevent)
{
  int     res;

  if (0 > gevent_serialize (gevent)) {
    fprintf (stderr, "tcp_send: could not serialize gevent\n");
    return -1;
  }

  if (1 != conn->is_connected) {
    fprintf (stderr, "tcp_send: can't write to closed socket\n");
    return -1;
  }

  if (0 > (res = write (conn->sock, gevent->generic.serial, gevent->generic.serial_len)) ) {
    fprintf (stderr, "tcp_send: failed to send event, %s\n", strerror(errno));
    return res;
  }

  conn->bytes_out += res;

  return res;
}

int
tcp_recv (TCPConn *conn, GEvent *gevent)
{
  int     len, res, flags = 0, s_1, s_2;
  char    numbuf[8];
  int     discard = 0;

  if (conn->so_timeout > 0)
    flags |= MSG_DONTWAIT;

  while (1) {
    if (conn->buf_leng >= 8 && conn->buf[7] == ':') {     /* the end of a message header? */
      memcpy (numbuf, conn->buf, 7);
      numbuf[7] = '\0';
      len = atoi (numbuf);
      if (len > 0) {
        if (len > TCP_BUFFER_SIZE) {
	  fprintf (stderr, "tcp_recv: incoming event too large - discarding [%d]\n", len);
	  if (conn->buf_leng > len + 8) {
            for (s_1 = 0, s_2 = len+8; s_1 < conn->buf_leng - (len+8) && s_2 < conn->buf_leng; s_1++, s_2++)
              conn->buf[s_1] = conn->buf[s_2];
            conn->buf_leng -= (len + 8);
            bzero (conn->buf + conn->buf_leng, TCP_BUFFER_SIZE - conn->buf_leng);
	    fprintf (stderr, "tcp_recv: [1] discarding %d bytes; 0 remaining\n", len + 8);
	  }
	  else {
	    discard = len - (conn->buf_leng - 8);
	    bzero (conn->buf, TCP_BUFFER_SIZE);
	    fprintf (stderr, "tcp_recv: [2] discarding %d bytes; %d remaining\n", conn->buf_leng, discard);
	    conn->buf_leng = 0;
	  }
          continue;
	}
        if (conn->buf_leng >= len + 8) {
	  gevent_deserialize (gevent, conn->buf, len+8);
          for (s_1 = 0, s_2 = len+8; s_1 < conn->buf_leng - (len+8) && s_2 < conn->buf_leng; s_1++, s_2++)
            conn->buf[s_1] = conn->buf[s_2];
          conn->buf_leng -= (len + 8);
          bzero (conn->buf + conn->buf_leng, TCP_BUFFER_SIZE - conn->buf_leng);
          return len;
        }
      }
    }

    if (32 <= TCP_BUFFER_SIZE - conn->buf_leng) {
      if (0 > (res = recv (conn->sock, conn->buf+conn->buf_leng, 32, flags)) ) {
        if (conn->so_timeout <= 0) {
          fprintf (stderr, "tcp_recv: an error/timeout occured in recv()\n");
          return -1;
        }
      }
      else if (0 == res) {
        if (conn->so_timeout <= 0)
          return -1;
      }
      else {
        conn->buf_leng += res;
        conn->bytes_in += res;
      }
    }

    if (discard > 0 && conn->buf_leng > 0) {
      if (conn->buf_leng > discard) {
        fprintf (stderr, "tcp_recv: [3] discarding %d bytes; 0 remaining\n", discard);
        for (s_1 = 0, s_2 = conn->buf_leng; s_1 < discard && s_2 < conn->buf_leng; s_1++, s_2++)
          conn->buf[s_1] = conn->buf[s_2];
        conn->buf_leng -= discard;
	discard = 0;
        bzero (conn->buf + conn->buf_leng, TCP_BUFFER_SIZE - conn->buf_leng);
      }
      else {
        discard -= conn->buf_leng;
        fprintf (stderr, "tcp_recv: [4] discarding %d bytes; %d remaining\n", conn->buf_leng, discard);
        conn->buf_leng = 0;
        bzero (conn->buf, TCP_BUFFER_SIZE);
        continue;
      }
    }

    usleep (100);
  }

  return 0;
}

int
udp_send (UDPConn *conn, UDPMesg *mesg, char *addr, int port)
{
  int        res;
  struct sockaddr_in  addr_info;
  socklen_t           addr_leng;
  char       addr_str[16];
  GEvent     *gevent;

  gevent = mesg->gevent;
  gevent_serialize (mesg->gevent);

  addr_leng = sizeof (addr_info);

  if (mesg->is_reply == 1) {
    memcpy (&addr_info, &mesg->info, addr_leng);
  }
  else {
    memcpy (&addr_info, &conn->sock_info, addr_leng);
    if (addr[0] != '\0')
      inet_pton (AF_INET, addr, &addr_info.sin_addr);

    if (port > 0)
      addr_info.sin_port = htons (port);
  }

  if (0 > (res = sendto (conn->sock, gevent->generic.serial,
                         gevent->generic.serial_len, 0, (SA *)&addr_info, addr_leng)) ) {
    fprintf (stderr, "udp_send: failed while doing sendto()\n");
    return -1;
  }

  inet_ntop (AF_INET, &addr_info.sin_addr, addr_str, 16);

  conn->bytes_out += gevent->generic.serial_len;

  return gevent->generic.serial_len;
}

void
udp_set_sotimeout (UDPConn *conn, int sotimeout)
{
  conn->so_timeout = sotimeout;
}

void
tcp_set_sotimeout (TCPConn *conn, int sotimeout)
{
  conn->so_timeout = sotimeout;
}

int
udp_recv (UDPConn *conn, UDPMesg *mesg)
{
  int        flags = 0, res = -2;
  char       *addr;

  struct timeval time_1, time_2;
  struct timezone tzone;

  if (conn == NULL || mesg == NULL)
    return -1;

  if (conn->so_timeout > 0)
    flags |= MSG_DONTWAIT;

  mesg->conn = conn;
  mesg->info_len = sizeof (mesg->info);

  /* repeat until we read a single valid message */
  while (1) {
    /* printf ("udp_recv: blocking on recvfrom()\n"); */

    if (conn->so_timeout > 0)
      gettimeofday (&time_1, &tzone);

    while (1) {
      if (0 > (res = recvfrom (conn->sock, conn->buf, UDP_BUFFER_SIZE,
                               flags, (SA *)&mesg->info, &mesg->info_len)) ) {
        if (conn->so_timeout <= 0) {
          break;
        }
        else {
          gettimeofday (&time_2, &tzone);

          if (time_2.tv_usec - time_1.tv_usec >= conn->so_timeout)
            break;
          else {
            usleep (1000);
            continue;
          }
        }
      }

      break;
    }

    if (res == 0) {
      fprintf (stderr, "udp_recv: recvfrom() returned 0. closed socket?\n");
      return -1;
    }
    else if (res == -1) {
      if (conn->so_timeout > 0) {
        gettimeofday (&time_2, &tzone);
        if (time_2.tv_usec - time_1.tv_usec >= conn->so_timeout)
          return -1;
      }
      continue;
    }
    else if (res == -2) {
      fprintf (stderr, "udp_recv: operation timed out while waiting on recvfrom()\n");
      usleep (1000);
      continue;
    }

    conn->bytes_in += res;
    mesg->port = ntohs (mesg->info.sin_port);
  
    if (NULL == ((const char *)addr = inet_ntop (AF_INET, &mesg->info.sin_addr, mesg->addr, 16)) ) {
      fprintf (stderr, "udp_recv: error while parsing client address\n");
      if (conn->so_timeout > 0) {
        gettimeofday (&time_2, &tzone);
        if (time_2.tv_usec - time_1.tv_usec >= conn->so_timeout)
          return -1;
      }
      continue;
    }

    /* printf ("udp_recv: read %.5d bytes from %s:%d\n", res, addr, mesg->port); */

    gevent_deserialize (mesg->gevent, conn->buf, res);

    return res;
  }

  return -1;
}

