#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <pthread.h>
#include <getopt.h>
#include <string.h>
#include <signal.h>

#include <X11/Xlib.h>
#include <X11/extensions/XTest.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>

#include "xgroove.h"
#include "grid.h"
#include "network.h"
#include "xgmaster.h"
#include "gx11.h"

int gmaster_abort = 0;

void
do_gmaster (int argc, char *argv[])
{
  GMaster     *gmaster;

  printf ("Initializing XGroove v1.0 - Master mode\n");

  signal (SIGPIPE, gmaster_sig_pipe);
  signal (SIGINT,  gmaster_sig_quit);
  signal (SIGTERM, gmaster_sig_quit);

  if (NULL == (gmaster = init_gmaster(argv[2])) ) {
    fprintf (stderr, "do_gmaster: failed to init_gmaster()\n");
    return;
  }

  while (1 && gmaster_abort == 0) {
    if (1 == gmaster->is_locked)
      gmaster_lock_mode (gmaster);
    else
      gmaster_free_mode (gmaster);
  }

  gmaster_setmode_free (gmaster);

  if (gmaster_abort != 0) {
    printf ("do_gmaster: killing threads...\n");
    pthread_kill (gmaster->tcp_lstn_thread, SIGTERM);
    pthread_kill (gmaster->udp_conn_thread, SIGTERM);
  }

  pthread_join (gmaster->tcp_lstn_thread, NULL);
  pthread_join (gmaster->udp_conn_thread, NULL);

  deinit_gmaster (gmaster);
}

GMaster *
init_gmaster (char *name)
{
  int        i, res;
  GMaster    *gmaster;

  if (NULL == (gmaster = malloc (sizeof(struct _GMaster))) ) {
    fprintf (stderr, "init_gmaster: failed doing malloc\n");
    return NULL;
  }

  for (i = 0; i < GRID_SIZE; i++) {
    gmaster->slaves[i] = NULL;
    gmaster->slvtmp[i] = NULL;
  }

  gmaster->name = name;
  gmaster->tcp_port = DEFAULT_TCP_PORT;
  gmaster->udp_port = DEFAULT_UDP_PORT;
  gmaster->grid = init_grid ();
  gmaster->main_screen = init_gridloc ();
  gmaster->main_screen->gslave = NULL;
  gmaster->ctrl_screen = gmaster->main_screen;
  gmaster->ctrl_slave = NULL;
  gmaster->clipboard = NULL;
  gmaster->clipboard_len = 0;                     /* size of malloc()'ed memory length */
  gmaster->gdisplay.display = NULL;               /* defined later with XOpenDisplay() */

  gridloc_setname (gmaster->main_screen, "master");
  
  while (NULL == (gmaster->tcp_lstn = tcp_open_listener (DEFAULT_TCP_PORT))) {
    fprintf (stderr, "init_gmaster: failed to start tcp listener on port %d; pausing...\n", DEFAULT_TCP_PORT);
    sleep (5);
  }

  if (NULL == (gmaster->udp_conn = udp_open_connection ("", DEFAULT_UDP_PORT)) ) {
    fprintf (stderr, "init_gmaster: failed to start udp listener on port %d\n", DEFAULT_UDP_PORT);
    deinit_gmaster (gmaster);
    return NULL;
  }

  grid_add (gmaster->grid, gmaster->main_screen);

  if (0 > (res = pthread_create (&gmaster->tcp_lstn_thread, NULL, (void *)&gmaster_tcp_accept_loop, gmaster)) ) {
    fprintf (stderr, "init_gmaster: tcp loop pthread_create failed, %s\n", strerror(res));
    deinit_gmaster (gmaster);
    return NULL;
  }

  if (0 > (res = pthread_create (&gmaster->udp_conn_thread, NULL, (void *)&gmaster_udp_accept_loop, gmaster)) ) {
    fprintf (stderr, "init_gmaster: udp loop pthread_create failed, %s\n", strerror(res));
    deinit_gmaster (gmaster);
    return NULL;
  }

  if (NULL == (gmaster->gdisplay.display = XOpenDisplay (NULL)) ) {
    fprintf (stderr, "init_gmaster: failed to connect to x server at %s\n", getenv("DISPLAY"));
    deinit_gmaster (gmaster);
    return NULL;
  }

  gmaster->gdisplay.width = DisplayWidth (gmaster->gdisplay.display, DefaultScreen(gmaster->gdisplay.display));
  gmaster->gdisplay.height = DisplayHeight (gmaster->gdisplay.display, DefaultScreen(gmaster->gdisplay.display));
  gmaster->gdisplay.rootwin = &DefaultRootWindow (gmaster->gdisplay.display);
  gmaster->gdisplay.mainwin = XCreateSimpleWindow (gmaster->gdisplay.display, *gmaster->gdisplay.rootwin, 
                                                   100, 100, 100, 100, 2, 1, 1);
  gmaster->gdisplay.pos_h = 100;
  gmaster->gdisplay.pos_v = 100;
  gmaster->ctrl_display = &gmaster->gdisplay;

  return gmaster;
}

void
deinit_gmaster (GMaster *gmaster)
{
  int     i;

  if (gmaster == NULL)
    return;

  tcp_close_listener (gmaster->tcp_lstn);
  udp_close_connection (gmaster->udp_conn);

  for (i = 0; i < GRID_SIZE; i++) {
    deinit_gslave (gmaster->slaves[i]);
    gmaster->slaves[i] = NULL;
  }

  if (gmaster->clipboard_len > 0) {
    free (gmaster->clipboard);
    gmaster->clipboard = NULL;
    gmaster->clipboard_len = 0;
  }

  if (gmaster->main_screen != NULL) {
    deinit_gridloc (gmaster->main_screen);
    gmaster->main_screen = NULL;
  }

  XCloseDisplay (gmaster->gdisplay.display);

  deinit_grid (gmaster->grid);

  free (gmaster);
}

void
gmaster_tcp_accept_loop (GMaster *gmaster)
{
  int     res;

  printf ("gmaster_tcp_accept_loop: start\n");

  while (1 && gmaster_abort == 0) {
    TCPConn     *conn;
    GSlave      *gslave;
    GridLoc     *gloc;
    GEvent      *gevent = init_gevent ();

    if (NULL == (conn = tcp_accept (gmaster->tcp_lstn)) ) {
      fprintf (stderr, "gmaster_tcp_accept_loop: failed to tcp_accept()\n");
      deinit_gevent (gevent);
      return;
    }

    printf ("gmaster_tcp_accept_loop: waiting for new-conn init event\n");

    if (0 > tcp_recv (conn, gevent)) {
      fprintf (stderr, "gmaster_tcp_accept_loop: failed to read init event\n");
      deinit_gevent (gevent);
      gevent = NULL;
      tcp_close_connection (conn);
      return;
    }

    if (gevent->generic.type != GrooveConfigureEvent) {
      fprintf (stderr, "gmaster_tcp_accept_loop: received invalid event: %d\n", gevent->generic.type);
      fprintf (stderr, "gmaster_tcp_accept_loop: closing connection to slave\n");
      deinit_gevent (gevent);
      gevent = NULL;
      tcp_close_connection (conn);
      continue;
    }

    if (NULL == (gslave = malloc (sizeof(struct _GSlave))) ) {
      fprintf (stderr, "gmaster_tcp_accept_loop: failed to malloc a gslave.\n");
      deinit_gevent (gevent);
      gevent = NULL;
      tcp_close_connection (conn);
      return;
    }

    gslave->gdisplay.width = gevent->gconfigure.width;
    gslave->gdisplay.height = gevent->gconfigure.height;

    printf ("gmaster_tcp_accept_loop: %.4d x %.4d\n", gslave->gdisplay.width, gslave->gdisplay.height);

    gslave->tcp_conn = conn;
    gslave->udp_conn = NULL;
    gloc = init_gridloc ();
    gridloc_setname (gloc, conn->addr);
    gloc->gslave = gslave;
    gslave->gdisplay.pos_h = 0;
    gslave->gdisplay.pos_v = 0;
    gslave->gdisplay.display = NULL;
    gslave->gmaster = gmaster;
    gslave->clipboard = NULL;
    gslave->clipboard_len = 0;
    gslave->is_controlled = 0;

    grid_add (gmaster->grid, gloc);

    if (0 > (res = pthread_create (&gslave->thread, NULL, (void *)&gmaster_tcp_slave_loop, gslave)) ) {
      fprintf (stderr, "gmaster_tcp_accept_loop: failed to spawn thread, %s\n", strerror(res));
      tcp_close_connection (conn);
      gslave->tcp_conn = NULL;
      deinit_gslave (gslave);
      gevent = NULL;
      continue;
    }

    deinit_gevent (gevent);
    gevent = NULL;
    printf ("gmaster_tcp_accept_loop: completed one loop\n");
  }

  printf ("gmaster_tcp_accept_loop: stop\n");
}

void
gmaster_close_conn (GMaster *gmaster, GSlave *gslave)
{
  GridLoc     *gloc;

  if (gslave->tcp_conn != NULL) {
    while (NULL != (gloc = grid_getbyname(gmaster->grid, gslave->tcp_conn->addr))) {
      grid_del (gmaster->grid, gloc);
      deinit_gridloc (gloc);
    }

    if (1 == gslave->is_controlled)
      gmaster->ctrl_slave = NULL;

    if (gslave->tcp_conn != NULL) {
      printf ("gmaster_close_conn: closing tcp connection\n");
      tcp_close_connection (gslave->tcp_conn);
      gslave->tcp_conn = NULL;
    }
  }

  printf ("gmaster_close_conn: freeing gslave\n");
  deinit_gslave (gslave);
}

void
gmaster_udp_accept_loop (GMaster *gmaster)
{
  int         res;
  UDPMesg     mesg;

  while (1 && gmaster_abort == 0) {
    mesg.gevent = init_gevent ();

    if (0 > (res = udp_recv (gmaster->udp_conn, &mesg))) {
      fprintf (stderr, "gmaster_udp_accept_loop: read 0 bytes - weird\n");
      deinit_gevent (mesg.gevent);
      sleep (1);
      continue;
    }

    printf ("gmaster_udp_accept_loop: read %d bytes from %s:%d\n", res, mesg.addr, mesg.port);

    if (mesg.gevent->generic.type != GrooveHelloEvent) {
      fprintf (stderr, "gmaster_udp_accept_loop: unknown event type: %d\n", mesg.gevent->generic.type);
      deinit_gevent (mesg.gevent);
      sleep (1);
      continue;
    }

    if (0 != strcmp (mesg.gevent->ghello.name, gmaster->name)) {
      fprintf (stderr, "gmaster_udp_accept_loop: ignoring, message for %s\n", mesg.gevent->ghello.name);
      deinit_gevent (mesg.gevent);
      sleep (1);
      continue;
    }

    printf ("gmaster_udp_accept_loop: received hello from %s ... sending reply\n", mesg.addr);

    mesg.gevent->ghelloback.type = GrooveHelloBackEvent;
    mesg.gevent->ghelloback.port = gmaster->tcp_lstn->port;
    mesg.is_reply = 1;

    udp_send (gmaster->udp_conn, &mesg, "", 0);
    deinit_gevent (mesg.gevent);
  }

  printf ("gmaster_udp_accept_loop: stop\n");
}

void
gmaster_tcp_slave_loop (void *arg)
{
  GSlave     *gslave = (GSlave *)arg;
  int        res;
  GMaster    *gmaster = (GMaster *)gslave->gmaster;

  printf ("gmaster_tcp_slave_loop: start\n");

  while (1) {
    GEvent     *gevent;

    if (NULL == (gevent = init_gevent ()) ) {
      fprintf (stderr, "gmaster_tcp_slave_loop: failed to init_gevent()\n");
      return;
    }

    if (0 > (res = tcp_recv (gslave->tcp_conn, gevent)) ) {
      fprintf (stderr, "gmaster_tcp_slave_loop: failed while doing tcp_recv(); closing down\n");
      deinit_gevent (gevent);
      return;
    }

    switch (gevent->generic.type) {
    case GrooveClipboardEvent:
      gmaster_proc_clipboard (gmaster, gevent->gclipboard);
      break;
    case GrooveGoodbyeEvent:
      printf ("gmaster_tcp_slave_loop: slave said goodbye [%s]\n", gslave->tcp_conn->addr);
      break;
    default:
      printf ("gmaster_tcp_slave_loop: Unknown event: %d\n", gevent->generic.type);
      break;
    }

    deinit_gevent (gevent);

    sleep (1);
  }
}

int
gmaster_move (GMaster *gmaster, int dir)
{
  GridLoc     *tmp_gloc;
  GEvent      *gevent;
  int         flags, ptr_state = 0;

  if (NULL == (tmp_gloc = grid_get (gmaster->grid, gmaster->ctrl_screen, dir)) ) {
    fprintf (stderr, "gmaster_move: could not move in direction %d\n", dir);
    return -1;
  }

  ptr_state = gx11_pointer_state_get (gmaster->gdisplay.display);

  /* if the a mouse button is pressed, don't change screens */
  if ((ptr_state & (Button1Mask | Button2Mask | Button3Mask | Button4Mask | Button5Mask)) != 0)
    return -1;

  flags = 0 | ButtonPressMask | ButtonReleaseMask | ButtonMotionMask | PointerMotionMask;

  /* the name of the new screen we're changing to */
  printf ("gmaster_move: tmp_gloc->name = %s\n", tmp_gloc->name);

  if (NULL == (gevent = init_gevent ()) ) {
    fprintf (stderr, "gmaster_move: failed to init_gevent()\n");
    return -1;
  }

  /* moving focus to a remote machine */
  if (tmp_gloc->gslave != NULL) {
    /* moving focus from the local machine */
    if (gmaster->ctrl_slave == NULL) {
      XGrabPointer (gmaster->gdisplay.display, *gmaster->gdisplay.rootwin, True,
                    flags, GrabModeAsync, GrabModeAsync, None, None, CurrentTime);
      XGrabKeyboard (gmaster->gdisplay.display, *gmaster->gdisplay.rootwin,
                     True, GrabModeAsync, GrabModeAsync, CurrentTime);
      XConvertSelection (gmaster->gdisplay.display, XA_PRIMARY, XA_STRING, XA_CUT_BUFFER0, 
                         gmaster->gdisplay.mainwin, CurrentTime);
      XFlush (gmaster->gdisplay.display);
    }

    if (0 == strcmp (gmaster->ctrl_screen->name, tmp_gloc->name)) {
      printf ("gmaster_move: already focused on this screen\n");
      deinit_gevent (gevent);
      return 0;
    }

    gevent->gcontrol.type = GrooveControlEvent;
    gevent->gcontrol.control_type = GrooveControlTypeSeize;

    if (0 > tcp_send (tmp_gloc->gslave->tcp_conn, gevent)) {
      fprintf (stderr, "gmaster_move: could not seize control; tcp_send failed\n");
      gmaster_close_conn (gmaster, tmp_gloc->gslave);
      gmaster_setmode_free (gmaster);
      deinit_gevent (gevent);
      return -1;
    }
    
    if (gmaster->ctrl_slave != NULL)
      gmaster->ctrl_slave->is_controlled = 0;

    gmaster->ctrl_screen = tmp_gloc;
    gmaster->ctrl_slave = tmp_gloc->gslave;
    gmaster->is_locked = 1;
    gmaster->ctrl_slave->is_controlled = 1;
  }
  /* moving focus to the local machine */
  else {
    /* moving focus from a remove machine */
    if (gmaster->ctrl_slave != NULL) {
      gevent->gcontrol.type = GrooveControlEvent;
      gevent->gcontrol.control_type = GrooveControlTypeCease;

      if (0 > tcp_send (gmaster->ctrl_slave->tcp_conn, gevent)) {
        fprintf (stderr, "gmaster_move: could not cease control; tcp_send failed\n");
        gmaster_close_conn (gmaster, gmaster->ctrl_slave);
        gmaster_setmode_free (gmaster);
        deinit_gevent (gevent);
        return -1;
      }

      gmaster->ctrl_slave->is_controlled = 0;

      XUngrabPointer (gmaster->gdisplay.display, CurrentTime);
      XUngrabKeyboard (gmaster->gdisplay.display, CurrentTime);
    }

    gmaster->ctrl_screen = tmp_gloc;
    gmaster_setmode_free (gmaster);

  }

  deinit_gevent (gevent);

  return 0;
}

void
gmaster_proc_motion (GMaster *gmaster, int pos_x, int pos_y)
{
  GEvent     *gevent = init_gevent ();
  GDisplay   *display = gmaster->ctrl_display;
  int        res = -1, tmp_x = 0, tmp_y = 0;

  gevent->gmotion.type = GrooveMotionEvent;

  if (1 == gmaster->is_locked) {
    tmp_x = gmaster->ctrl_display->pos_h + (pos_x - LOCKPOS_H);
    tmp_y = gmaster->ctrl_display->pos_v + (pos_y - LOCKPOS_V);

    tmp_x = tmp_x < 0 ? 0 : tmp_x;
    tmp_y = tmp_y < 0 ? 0 : tmp_y;
    tmp_x = tmp_x >= gmaster->ctrl_display->width  ? gmaster->ctrl_display->width  - 1 : tmp_x;
    tmp_y = tmp_y >= gmaster->ctrl_display->height ? gmaster->ctrl_display->height - 1 : tmp_y;
  }
  else {
    tmp_x = pos_x;
    tmp_y = pos_y;
  }

  gmaster->ctrl_display->pos_h = tmp_x;
  gmaster->ctrl_display->pos_v = tmp_y;

  if (display->pos_h <= 0) {                                         /* move left     */
    if (0 == (res = gmaster_move (gmaster, SCREEN_LEFT))) {
      tmp_x = gmaster->ctrl_display->width - 2;
      tmp_y = (int)(display->pos_v / (gmaster->ctrl_display->height / display->height));
    }
  }
  else if (display->pos_h >= display->width - 1) {                   /* move right    */
    if (0 == (res = gmaster_move (gmaster, SCREEN_RIGHT))) {
      tmp_x = 1;
      tmp_y = (int)(display->pos_v / (gmaster->ctrl_display->height / display->height));
    }
  }
  else if (0 >= display->pos_v) {                                    /* move up       */
    if (0 == (res = gmaster_move (gmaster, SCREEN_UP))) {
      tmp_x = (int)(display->pos_h / (gmaster->ctrl_display->width / display->width));
      tmp_y = display->height - 2;
    }
  }
  else if (display->pos_v >= display->height - 1) {                /* move down     */
    if (0 == (res = gmaster_move (gmaster, SCREEN_DOWN))) {
      tmp_x = (int)(display->pos_h / (gmaster->ctrl_display->width / display->width));
      tmp_y = 1;
    }
  }

  if (res == 0 && gmaster->is_locked == 1) {
    printf ("gmaster_loop: [1] initializing new screen coordinates [%.4d x %.4d]\n", tmp_x, tmp_y);
    gevent->gmotion.x = tmp_x;
    gevent->gmotion.y = tmp_y;
    gevent->gmotion.motion_type = GrooveMotionTypeAbsolute;

    if (0 > tcp_send (gmaster->ctrl_slave->tcp_conn, gevent)) {
      fprintf (stderr, "gmaster_proc_motion: could not continue; tcp_send failed [1]\n");
      gmaster_close_conn (gmaster, gmaster->ctrl_slave);
      gmaster_setmode_free (gmaster);
      return;
    }

    gmaster->ctrl_display->pos_h = tmp_x;
    gmaster->ctrl_display->pos_v = tmp_y;

    XTestFakeMotionEvent (gmaster->gdisplay.display, None, LOCKPOS_H, LOCKPOS_V, 0);
    XFlush (gmaster->gdisplay.display);
  }
  else if (res == 0 && gmaster->is_locked != 1) {
    printf ("gmaster_loop: [2] initializing new screen coordinates [%.4d x %.4d]\n", tmp_x, tmp_y);
    gmaster->ctrl_display->pos_h = tmp_x;
    gmaster->ctrl_display->pos_v = tmp_y;
    XTestFakeMotionEvent (gmaster->gdisplay.display, None, tmp_x, tmp_y, 0);
    XFlush (gmaster->gdisplay.display);
  }
  else if (res != 0 && gmaster->is_locked == 1) {
    gevent->gmotion.x = gmaster->ctrl_display->pos_h;
    gevent->gmotion.y = gmaster->ctrl_display->pos_v;
    gevent->gmotion.motion_type = GrooveMotionTypeAbsolute;

    if (0 > tcp_send (gmaster->ctrl_slave->tcp_conn, gevent)) {
      fprintf (stderr, "gmaster_proc_motion: failed to send motion event; tcp_send failed [2]\n");
      gmaster_close_conn (gmaster, gmaster->ctrl_slave);
      gmaster_setmode_free (gmaster);
      return;
    }

    XTestFakeMotionEvent (gmaster->gdisplay.display, None, LOCKPOS_H, LOCKPOS_V, 0);
    XFlush (gmaster->gdisplay.display);
  }
}

void
gmaster_proc_click (GMaster *gmaster, int btn, int is_press)
{
  GEvent     *gevent = init_gevent ();

  gevent->gclick.type = GrooveClickEvent;
  gevent->gclick.button = btn;

  if (1 != gmaster->is_locked)
    return;

  if (1 == is_press)
    gevent->gclick.click_type = GrooveClickTypePress;
  else
    gevent->gclick.click_type = GrooveClickTypeRelease;

  if (0 > tcp_send (gmaster->ctrl_slave->tcp_conn, gevent)) {
    fprintf (stderr, "gmaster_proc_click: failed to send click; tcp_send failed\n");
    gmaster_close_conn (gmaster, gmaster->ctrl_slave);
    gmaster_setmode_free (gmaster);
    return;
  }
}

void
gmaster_xevent_selection_send (gmaster, sre)
  GMaster *gmaster;
  XSelectionRequestEvent sre;
{
  gx11_proc_selection_send (gmaster->clipboard, gmaster->clipboard_len, sre);
}

void
gmaster_xevent_selection_notify (gmaster, window)
  GMaster *gmaster;
  Window window;
{
  GEvent     *gevent = init_gevent ();
  Display    *display = gmaster->gdisplay.display;

  if (gmaster->clipboard_len > 0) {
    free (gmaster->clipboard);
    gmaster->clipboard = NULL;
    gmaster->clipboard_len = 0;
  }

  gmaster->clipboard = gx11_proc_selection_notify (display, window, &gmaster->clipboard_len);

  if (NULL == (gevent->gclipboard.clipboard = malloc (gmaster->clipboard_len))) {
    fprintf (stderr, "gmaster_xevent_selection_notify: failed in malloc(%d)\n", gmaster->clipboard_len);
    deinit_gevent (gevent);
    return;
  }

  memcpy (gevent->gclipboard.clipboard, gmaster->clipboard, gmaster->clipboard_len);

  gevent->gclipboard.type = GrooveClipboardEvent;
  gevent->gclipboard.clipboard_len = gmaster->clipboard_len;

  tcp_send (gmaster->ctrl_slave->tcp_conn, gevent);
}

void
gmaster_proc_keypress (GMaster *gmaster, int keycode, int is_press)
{
  GEvent     *gevent;

  if (1 != gmaster->is_locked)
    return;

  gevent = init_gevent ();
  gevent->gkeypress.type = GrooveKeyPressEvent;
  gevent->gkeypress.keycode = keycode;

  if (1 == is_press)
    gevent->gkeypress.keypress_type = GrooveKeyPressTypePress;
  else
    gevent->gkeypress.keypress_type = GrooveKeyPressTypeRelease;

  if (0 > tcp_send (gmaster->ctrl_slave->tcp_conn, gevent)) {
    fprintf (stderr, "gmaster_proc_keypress: failed to send keypress; tcp_send failed\n");
    gmaster_close_conn (gmaster, gmaster->ctrl_slave);
    gmaster_setmode_free (gmaster);
    deinit_gevent (gevent);
    return;
  }

  deinit_gevent (gevent);
}

void
gmaster_proc_clipboard (gmaster, gce)
  GMaster *gmaster;
  GClipboardEvent gce;
{
  Display     *display = gmaster->gdisplay.display;
  Window      mainwin = gmaster->gdisplay.mainwin;

  if (gmaster->clipboard_len > 0) {
    free (gmaster->clipboard);
    gmaster->clipboard = NULL;
    gmaster->clipboard_len = 0;
  }

  gmaster->clipboard = gx11_proc_clipboard (display, mainwin, gce.clipboard, gce.clipboard_len);
  gmaster->clipboard_len = gce.clipboard_len;
}

void
gmaster_free_mode (GMaster *gmaster)
{
  Display     *display = gmaster->gdisplay.display;
  XEvent      event;
  int         prev_x = -1, prev_y = -1;
  int         tmp_x = -1, tmp_y = -1;

  while (1 != gmaster->is_locked) {
    if (1 <= XPending (gmaster->gdisplay.display)) {
      XNextEvent (gmaster->gdisplay.display, &event);

      switch (event.type) {
      case SelectionNotify:
        gmaster_xevent_selection_notify (gmaster, event.xselection.requestor);
        break;
      case SelectionRequest:
        gmaster_xevent_selection_send (gmaster, event.xselectionrequest);
        break;
      case SelectionClear:
        printf ("gmaster_lock_mode: SelectionClear\n");
        break;
      default:
        break;
      }
    }

    gx11_pointer_pos_get (display, &tmp_x, &tmp_y);

    if( tmp_x == prev_x && tmp_y == prev_y) {
      usleep (1000);
      continue;
    }

    prev_x = tmp_x;
    prev_y = tmp_y;

    gmaster_proc_motion (gmaster, prev_x, prev_y);

    usleep (1000);
  }

}

void
gmaster_lock_mode (GMaster *gmaster)
{
  XEvent     xevent;

  while (1 == gmaster->is_locked)
  {
    if (1 > XPending (gmaster->gdisplay.display)) {
      usleep (1000);
      continue;
    }

    XNextEvent (gmaster->gdisplay.display, &xevent);

    switch (xevent.type) {
    case MotionNotify:
      if (xevent.xmotion.x == LOCKPOS_H && xevent.xmotion.y == LOCKPOS_V)
        break;
      gmaster_proc_motion (gmaster, xevent.xmotion.x, xevent.xmotion.y);
      break;
    case ButtonPress:
      gmaster_proc_click (gmaster, xevent.xbutton.button, 1);
      break;
    case ButtonRelease:
      gmaster_proc_click (gmaster, xevent.xbutton.button, 0);
      break;
    case KeyPress:
      gmaster_proc_keypress (gmaster, xevent.xkey.keycode, 1);
      break;
    case KeyRelease:
      gmaster_proc_keypress (gmaster, xevent.xkey.keycode, 0);
      break;
    case SelectionNotify:
      gmaster_xevent_selection_notify (gmaster, xevent.xselection.requestor);
      break;
    case SelectionRequest:
      gmaster_xevent_selection_send (gmaster, xevent.xselectionrequest);
      break;
    case SelectionClear:
      printf ("gmaster_lock_mode: SelectionClear\n");
      break;
    default:
      break;
    }
  }
}

void
gmaster_setmode_free (GMaster *gmaster)
{
  if (1 == gmaster->is_locked) {
    XUngrabPointer (gmaster->gdisplay.display, CurrentTime);
    XUngrabKeyboard (gmaster->gdisplay.display, CurrentTime);
  }

  gmaster->ctrl_slave = NULL;
  gmaster->is_locked = 0;
}

void
gmaster_sig_pipe (int signo)
{
  fprintf (stderr, "gmaster_sig_pipe: received signal\n");
}

void
gmaster_sig_quit (int signo)
{
  fprintf (stderr, "gmaster_sig_quit: received signal\n");

  if (gmaster_abort > 0)
    exit (0);

  gmaster_abort = 1;
}

